use std::collections::HashMap;

use serde::Serialize;

use crate::convert::map_lch_to_srgb_hex;

use super::{color_set::M3ColorSet, surface::M3SurfaceSet, tonal_palette::TonalPalette};

#[derive(Debug, Clone, Serialize)]
pub struct M3BaselineColors {
    pub primary: M3ColorSet,
    pub secondary: M3ColorSet,
    pub tertiary: M3ColorSet,
    pub error: M3ColorSet,
    pub surface: M3SurfaceSet,
    pub outline: String,
    pub outline_variant: String,
    pub scrim: String,
    pub shadow: String,
    pub customs: HashMap<String, M3ColorSet>,
    dark_set: bool,
}

impl M3BaselineColors {
    pub fn new(
        p: &TonalPalette,
        s: &TonalPalette,
        t: &TonalPalette,
        n: &TonalPalette,
        nv: &TonalPalette,
        e: &TonalPalette,
        dark_set: bool,
    ) -> Self {
        let color_set_generator = if dark_set {
            M3ColorSet::new_dark_set
        } else {
            M3ColorSet::new_light_set
        };
        let surface_set_generator = if dark_set {
            M3SurfaceSet::new_dark_set
        } else {
            M3SurfaceSet::new_light_set
        };

        let primary = color_set_generator(p);
        let secondary = color_set_generator(s);
        let tertiary = color_set_generator(t);
        let surface = surface_set_generator(n, nv);
        let error = color_set_generator(e);
        let outline = if dark_set { n.tone(60.0) } else { n.tone(50.0) };
        let outline_variant = if dark_set {
            nv.tone(30.0)
        } else {
            nv.tone(80.0)
        };
        let scrim = n.tone(0.0);
        let shadow = n.tone(0.0);

        Self {
            primary,
            secondary,
            tertiary,
            error,
            surface,
            outline: map_lch_to_srgb_hex(&outline),
            outline_variant: map_lch_to_srgb_hex(&outline_variant),
            scrim: map_lch_to_srgb_hex(&scrim),
            shadow: map_lch_to_srgb_hex(&shadow),
            customs: HashMap::new(),
            dark_set,
        }
    }

    pub fn add_custom_set(&mut self, name: String, c: &TonalPalette) {
        let color_set_generator = if self.dark_set {
            M3ColorSet::new_dark_set
        } else {
            M3ColorSet::new_light_set
        };
        self.customs.insert(name, color_set_generator(c));
    }

    pub fn full_custom(
        primary: M3ColorSet,
        secondary: M3ColorSet,
        tertiary: M3ColorSet,
        error: M3ColorSet,
        surface: M3SurfaceSet,
        outline: String,
        outline_variant: String,
        scrim: String,
        shadow: String,
        customs: HashMap<String, M3ColorSet>,
        dark_set: bool,
    ) -> Self {
        Self {
            primary,
            secondary,
            tertiary,
            error,
            surface,
            outline,
            outline_variant,
            scrim,
            shadow,
            customs,
            dark_set,
        }
    }

    pub fn to_css_variables(&self) -> Vec<String> {
        let mut css_variables = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        css_variables.extend(self.primary.to_css_variables(prefix, "primary"));
        css_variables.extend(self.secondary.to_css_variables(prefix, "secondary"));
        css_variables.extend(self.tertiary.to_css_variables(prefix, "tertiary"));
        css_variables.extend(self.error.to_css_variables(prefix, "error"));
        css_variables.extend(self.surface.to_css_variables(prefix));
        css_variables.push(format!("--color-{}-outline: #{};", prefix, self.outline));
        css_variables.push(format!(
            "--color-{}-outline-variant: #{};",
            prefix, self.outline_variant
        ));
        css_variables.push(format!("--color-{}-scrim: #{};", prefix, self.scrim));
        css_variables.push(format!("--color-{}-shadow: #{};", prefix, self.shadow));
        for (name, color_set) in &self.customs {
            css_variables.extend(color_set.to_css_variables(prefix, &name.to_lowercase()));
        }

        css_variables
    }

    pub fn to_scss_variables(&self) -> Vec<String> {
        let mut scss_variables = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        scss_variables.extend(self.primary.to_scss_variables(prefix, "primary"));
        scss_variables.extend(self.secondary.to_scss_variables(prefix, "secondary"));
        scss_variables.extend(self.tertiary.to_scss_variables(prefix, "tertiary"));
        scss_variables.extend(self.error.to_scss_variables(prefix, "error"));
        scss_variables.extend(self.surface.to_scss_variables(prefix));
        scss_variables.push(format!("$color-{}-outline: #{};", prefix, self.outline));
        scss_variables.push(format!(
            "$color-{}-outline-variant: #{};",
            prefix, self.outline_variant
        ));
        scss_variables.push(format!("$color-{}-scrim: #{};", prefix, self.scrim));
        scss_variables.push(format!("$color-{}-shadow: #{};", prefix, self.shadow));
        for (name, color_set) in &self.customs {
            scss_variables.extend(color_set.to_scss_variables(prefix, &name.to_lowercase()));
        }

        scss_variables
    }

    pub fn to_javascript_object_fields(&self) -> Vec<String> {
        let mut js_object_fields = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        js_object_fields.extend(self.primary.to_javascript_object_fields(prefix, "primary"));
        js_object_fields.extend(
            self.secondary
                .to_javascript_object_fields(prefix, "secondary"),
        );
        js_object_fields.extend(
            self.tertiary
                .to_javascript_object_fields(prefix, "tertiary"),
        );
        js_object_fields.extend(self.error.to_javascript_object_fields(prefix, "error"));
        js_object_fields.extend(self.surface.to_javascript_object_fields(prefix));
        js_object_fields.push(format!("{}Outline: '#{}',", prefix, self.outline));
        js_object_fields.push(format!(
            "{}OutlineVariant: '#{}',",
            prefix, self.outline_variant
        ));
        js_object_fields.push(format!("{}Scrim: '#{}',", prefix, self.scrim));
        js_object_fields.push(format!("{}Shadow: '#{}',", prefix, self.shadow));
        for (name, color_set) in &self.customs {
            js_object_fields
                .extend(color_set.to_javascript_object_fields(prefix, &name.to_lowercase()));
        }

        js_object_fields
    }
}
