use std::rc::Rc;

use palette::Lch;

use crate::schemes::material_design_3::TonalPalette;

use super::{contrast_curve::ContrastCurve, dynamic_scheme::DynamicScheme};

pub type TonalPaletteGenerator = Box<dyn Fn(&DynamicScheme) -> TonalPalette>;
pub type ToneSearcher = Box<dyn Fn(&DynamicScheme) -> f32>;
pub type DynamicColorSearcher = Box<dyn Fn(&DynamicScheme) -> Rc<DynamicColor>>;
pub type ToneDeltaPairGenerator = Box<dyn Fn(&DynamicScheme) -> ToneDeltaPair>;
pub type CustomPaletteGenerator = Box<dyn Fn(&Lch) -> TonalPalette>;

#[derive(Debug, Clone, Copy, PartialEq)]
#[allow(dead_code)]
pub enum TonePolarity {
    Darker,
    Lighter,
    Nearer,
    Farther,
}

pub struct DynamicColor {
    name: String,
    palette: TonalPaletteGenerator,
    tone: ToneSearcher,
    is_background: Option<bool>,
    background: Option<DynamicColorSearcher>,
    secondary_background: Option<DynamicColorSearcher>,
    contrast_curve: Option<ContrastCurve>,
    tone_delta_pairs: Option<ToneDeltaPairGenerator>,
}

pub struct ToneDeltaPair {
    pub role_a: Rc<DynamicColor>,
    pub role_b: Rc<DynamicColor>,
    pub delta: f32,
    pub polarity: TonePolarity,
    pub togather: bool,
}

impl DynamicColor {
    pub fn new(
        name: Option<&str>,
        palette: TonalPaletteGenerator,
        tone: ToneSearcher,
        is_background: Option<bool>,
        background: Option<DynamicColorSearcher>,
        secondary_background: Option<DynamicColorSearcher>,
        contrast_curve: Option<ContrastCurve>,
        tone_delta_pairs: Option<ToneDeltaPairGenerator>,
    ) -> Self {
        DynamicColor {
            name: name.unwrap_or("").to_string(),
            palette,
            tone,
            is_background: is_background.or(Some(false)),
            background,
            secondary_background,
            contrast_curve,
            tone_delta_pairs,
        }
    }

    pub fn tone(&self, scheme: &DynamicScheme) -> f32 {
        (self.tone)(scheme)
    }

    pub fn get_lch(&self, scheme: &DynamicScheme) -> Lch {
        let tone = self.get_tone(scheme);
        (self.palette)(scheme).tone(tone)
    }

    pub fn get_tone(&self, scheme: &DynamicScheme) -> f32 {
        let decreasing_contrast = scheme.contrast_level < 0.0;

        if let Some(pair_generator) = &self.tone_delta_pairs {
            let tone_delta = pair_generator(scheme);
            let bg = (self.background.as_ref().unwrap())(scheme);
            let bg_tone = bg.get_tone(scheme);
            let is_nearer = tone_delta.polarity == TonePolarity::Nearer
                || (tone_delta.polarity == TonePolarity::Lighter && !scheme.is_dark)
                || (tone_delta.polarity == TonePolarity::Darker && scheme.is_dark);
            let (nearer, farther) = if is_nearer {
                (&tone_delta.role_a, &tone_delta.role_b)
            } else {
                (&tone_delta.role_b, &tone_delta.role_a)
            };
            let expansion_factor = if scheme.is_dark { 1.0 } else { -1.0 };

            let n_contrast = (nearer.contrast_curve.as_ref().unwrap()).get(scheme.contrast_level);
            let f_contrast = (farther.contrast_curve.as_ref().unwrap()).get(scheme.contrast_level);

            let n_initial_tone = nearer.tone(scheme);
            let mut n_tone =
                if super::contrast::ratio_of_tones(bg_tone, n_initial_tone) >= n_contrast {
                    n_initial_tone
                } else {
                    foreground_tone(bg_tone, n_contrast)
                };
            let f_initial_tone = farther.tone(scheme);
            let mut f_tone =
                if super::contrast::ratio_of_tones(bg_tone, f_initial_tone) >= f_contrast {
                    f_initial_tone
                } else {
                    foreground_tone(bg_tone, f_contrast)
                };

            if decreasing_contrast {
                n_tone = foreground_tone(bg_tone, n_contrast);
                f_tone = foreground_tone(bg_tone, f_contrast);
            }

            if (f_tone - n_tone) * expansion_factor < tone_delta.delta {
                f_tone = (n_tone + tone_delta.delta * expansion_factor).clamp(0.0, 100.0);
            }

            if n_tone >= 50.0 && n_tone < 60.0 {
                if expansion_factor > 0.0 {
                    n_tone = 60.0;
                    f_tone = (n_tone + tone_delta.delta * expansion_factor).max(f_tone);
                } else {
                    n_tone = 49.0;
                    f_tone = (n_tone + tone_delta.delta * expansion_factor).min(f_tone);
                }
            } else if f_tone >= 50.0 && f_tone < 60.0 {
                if tone_delta.togather {
                    if expansion_factor > 0.0 {
                        n_tone = 60.0;
                        f_tone = (n_tone + tone_delta.delta * expansion_factor).max(f_tone);
                    } else {
                        n_tone = 49.0;
                        f_tone = (n_tone + tone_delta.delta * expansion_factor).min(f_tone);
                    }
                } else {
                    if expansion_factor > 0.0 {
                        f_tone = 60.0;
                    } else {
                        f_tone = 49.0;
                    }
                }
            }

            if self.name.eq_ignore_ascii_case(&nearer.name) {
                n_tone
            } else {
                f_tone
            }
        } else {
            let mut result = (self.tone)(scheme);

            if self.background.is_none() {
                return result;
            }

            let bg_tone = (self.background.as_ref().unwrap())(scheme).get_tone(scheme);
            let desired_ratio = (self.contrast_curve.as_ref().unwrap()).get(scheme.contrast_level);

            if super::contrast::ratio_of_tones(bg_tone, result) < desired_ratio {
                result = foreground_tone(bg_tone, desired_ratio);
            }
            if decreasing_contrast {
                result = foreground_tone(bg_tone, desired_ratio);
            }

            if self.is_background.unwrap_or(false) && result >= 50.0 && result < 60.0 {
                if super::contrast::ratio_of_tones(49.0, bg_tone) >= desired_ratio {
                    result = 49.0;
                } else {
                    result = 60.0;
                }
            }

            if let Some(secondary_background) = &self.secondary_background {
                let (bg_tone_1, bg_tone_2) = (
                    (self.background.as_ref().unwrap())(scheme).get_tone(scheme),
                    secondary_background(scheme).get_tone(scheme),
                );
                let (upper, lower) = (bg_tone_1.max(bg_tone_2), bg_tone_1.min(bg_tone_2));

                if super::contrast::ratio_of_tones(upper, result) >= desired_ratio
                    && super::contrast::ratio_of_tones(lower, result) >= desired_ratio
                {
                    return result;
                }

                let light_option = super::contrast::lighter(upper, desired_ratio);
                let dark_option = super::contrast::darker(lower, desired_ratio);

                let mut availables = vec![];
                if light_option != -1.0 {
                    availables.push(light_option);
                }
                if dark_option != -1.0 {
                    availables.push(dark_option);
                }

                if prefer_light_foreground(bg_tone_1) || prefer_light_foreground(bg_tone_2) {
                    return if light_option < 0.0 {
                        100.0
                    } else {
                        light_option
                    };
                }
                if availables.len() == 1 {
                    return availables[0];
                }

                return if dark_option < 0.0 { 0.0 } else { dark_option };
            }

            result
        }
    }
}

#[inline]
fn prefer_light_foreground(tone: f32) -> bool {
    tone.round() < 60.0
}

pub fn foreground_tone(background_tone: f32, ratio: f32) -> f32 {
    let lighter_tone = super::contrast::unsafe_lighter(background_tone, ratio);
    let darker_tone = super::contrast::unsafe_darker(background_tone, ratio);
    let lighter_ratio = super::contrast::ratio_of_tones(lighter_tone, background_tone);
    let darker_ratio = super::contrast::ratio_of_tones(darker_tone, background_tone);

    if prefer_light_foreground(background_tone) {
        let difference = (lighter_ratio - darker_ratio).abs() < 0.1
            && lighter_ratio < ratio
            && darker_ratio < ratio;
        if lighter_ratio >= ratio || lighter_ratio >= darker_ratio || difference {
            lighter_tone
        } else {
            darker_tone
        }
    } else {
        if darker_ratio >= ratio || darker_ratio >= lighter_ratio {
            darker_tone
        } else {
            lighter_tone
        }
    }
}
